import argparse
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from tqdm import tqdm
import wandb

from torchmetrics.classification import Accuracy, Precision, Recall, F1Score
from torchvision import models
from torchvision.models import ResNet18_Weights
from torchvision import transforms

import numpy as np
from torch.utils.data import Dataset
from PIL import Image

class CustomNPYDataset(Dataset):
    def __init__(self, data_file, labels_file, transform=None):
        self.images = np.load(data_file, allow_pickle=True)
        self.labels = np.load(labels_file)
        self.transform = transform

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        img_array = self.images[idx]
        img = Image.fromarray(img_array)
        label = self.labels[idx]
        if self.transform:
            img = self.transform(img)
        return img, label


# -----------------------------
# 🎯 Parse command-line arguments
# -----------------------------
parser = argparse.ArgumentParser()
parser.add_argument('--lr', type=float, default=0.001, help='Learning rate')
parser.add_argument('--batch_size', type=int, default=64, help='Batch size')
parser.add_argument('--epochs', type=int, default=10, help='Number of epochs')
args = parser.parse_args()

# -----------------------------
# 🚀 Initialize Weights & Biases
# -----------------------------
wandb.init(project="cifar10-mlp",entity="mpennisi", config=args)
config = wandb.config

# -----------------------------
# 📦 Setup
# -----------------------------
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
# Load pretrained ResNet18 using the weights API
model = models.resnet18(weights=ResNet18_Weights.DEFAULT)

# Check the number of input features for the classifier
in_features = model.fc.in_features

# Replace the final layer with a new one for 10 CIFAR classes
model.fc = nn.Linear(in_features, 10)

loss_fn = nn.CrossEntropyLoss()
optimizer = torch.optim.Adam(model.parameters(), lr=config.lr)

# -----------------------------
# 📊 Data Loaders
# -----------------------------
# Training transforms: with augmentation
train_transform = transforms.Compose([
    transforms.RandomHorizontalFlip(),
    transforms.RandomRotation(15),
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.5]*3, std=[0.5]*3)
])

# Test transforms: no augmentation
test_transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.5]*3, std=[0.5]*3)
])

train_dataset = CustomNPYDataset(
    data_file="./data/CIFAR10/train_data.npy",
    labels_file="./data/CIFAR10/train_labels.npy",
    transform=train_transform
)

test_dataset = CustomNPYDataset(
    data_file="./data/CIFAR10/test_data.npy",
    labels_file="./data/CIFAR10/test_labels.npy",
    transform=test_transform
)
train_loader = DataLoader(train_dataset, batch_size=config.batch_size, shuffle=True)
test_loader = DataLoader(test_dataset, batch_size=config.batch_size)

dataloaders = {
    'train': train_loader,
    'val': test_loader
}

# -----------------------------
# 📈 Define metrics per phase
# -----------------------------
metrics = {
    'train': {
        'accuracy': Accuracy(task="multiclass", num_classes=10).to(device),
        'precision': Precision(task="multiclass", num_classes=10, average='macro').to(device),
        'recall': Recall(task="multiclass", num_classes=10, average='macro').to(device),
        'f1': F1Score(task="multiclass", num_classes=10, average='macro').to(device),
    },
    'val': {
        'accuracy': Accuracy(task="multiclass", num_classes=10).to(device),
        'precision': Precision(task="multiclass", num_classes=10, average='macro').to(device),
        'recall': Recall(task="multiclass", num_classes=10, average='macro').to(device),
        'f1': F1Score(task="multiclass", num_classes=10, average='macro').to(device),
    }
}

# -----------------------------
# 🔁 Training Loop
# -----------------------------

model = model.to(device)

for epoch in range(config.epochs):
    print(f"\nEpoch {epoch + 1}/{config.epochs}")

    for phase in ['train', 'val']:
        model.train() if phase == 'train' else model.eval()

        running_loss = 0.0
        total_samples = 0

        for metric in metrics[phase].values():
            metric.reset()

        loop = tqdm(dataloaders[phase], desc=f"{phase.upper()} Phase", leave=False)

        for inputs, labels in loop:
            inputs, labels = inputs.to(device), labels.to(device)

            with torch.set_grad_enabled(phase == 'train'):
                outputs = model(inputs)
                loss = loss_fn(outputs, labels)

                if phase == 'train':
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()

            batch_size = inputs.size(0)
            running_loss += loss.item() * batch_size
            total_samples += batch_size

            preds = outputs.argmax(dim=1)
            for metric in metrics[phase].values():
                metric.update(preds, labels)

            loop.set_postfix(loss=loss.item(), acc=100.0 * metrics[phase]['accuracy'].compute().item())

        # Epoch summary
        epoch_loss = running_loss / total_samples
        epoch_metrics = {name: m.compute().item() for name, m in metrics[phase].items()}

        print(
            f"{phase.capitalize()} Loss: {epoch_loss:.4f}  "
            f"Acc: {epoch_metrics['accuracy']:.4f}  "
            f"Prec: {epoch_metrics['precision']:.4f}  "
            f"Recall: {epoch_metrics['recall']:.4f}  "
            f"F1: {epoch_metrics['f1']:.4f}"
        )

        # 🔍 Log to W&B
        wandb.log({
            f"{phase}_loss": epoch_loss,
            f"{phase}_accuracy": epoch_metrics['accuracy'],
            f"{phase}_precision": epoch_metrics['precision'],
            f"{phase}_recall": epoch_metrics['recall'],
            f"{phase}_f1": epoch_metrics['f1'],
            "epoch": epoch + 1
        })

        # 💾 Save checkpoint
        if phase == 'train' and epoch % 2 == 0:
            checkpoint = {
                'epoch': epoch,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
            }
            torch.save(checkpoint, f"checkpoint_epoch_{epoch}.pth")
